#!/usr/bin/python3
import json
import netaddr
import gzip
import re
import os
import sys
import pprint
import requests
import argparse
import datetime
import logging
import time

ap = argparse.ArgumentParser()
ap.add_argument("--raw", type=argparse.FileType("r", encoding="utf-8"), required=True)
ap.add_argument("--json", default=None, type=argparse.FileType('w'), help="Output a JSON file instead of printing to stdout")
ap.add_argument("--debug", action="store_true", help="Output tons of debug messages")
args = ap.parse_args()

if args.debug:
    logging.basicConfig(level=logging.DEBUG)

def filterNonAscii(s):
    return s.encode("ascii", errors="ignore").decode()

def getHardwareModelFromEntry(d):
    try: 
        hardware_model = d["nodeinfo"]["hardware"]["model"]
    except KeyError:
        hardware_model = "UNKNOWN"

    return hardware_model

def ipv62mac(ipv6):
    # remove subnet info if given
    subnetIndex = ipv6.find("/")
    if subnetIndex != -1:
        ipv6 = ipv6[:subnetIndex]

    ipv6Parts = ipv6.split(":")
    macParts = []
    for ipv6Part in ipv6Parts[-4:]:
        while len(ipv6Part) < 4:
            ipv6Part = "0" + ipv6Part
        macParts.append(ipv6Part[:2])
        macParts.append(ipv6Part[-2:])

    # modify parts to match MAC value
    macParts[0] = "%02x" % (int(macParts[0], 16) ^ 2)
    del macParts[4]
    del macParts[3]

    return ":".join(macParts)

class LogRecordParseError(Exception):
    def __init__(self, record_line):
        self.record_line = record_line

class LogRecord:
    def __init__(self,record):
        record = record.replace("%2B","+")
        regex = r'(fd21.*) - - \[(.*) "GET /gluon/(.*)/sysupgrade/gluon-ffs-([0-9]\.[0-9]\+[0-9]+-[0-9]+-[0-9]+-g\.[0-9a-z]+-s\.[0-9a-z]+)-(.*)\ HTTP/1.1" ([0-9]+) [0-9]+ "-" "(.*)"'
        result = re.search(regex,record)
        try:
            groups = result.groups()
        except:
            raise LogRecordParseError(record)
        self.ipv6 = groups[0]
        self.date = groups[1]
        self.branch = groups[2]
        self.release = groups[3]
        self.model = groups[4]
        self.status = groups[5]
        self.agent = groups[6]
        self.segment = int(self.ipv6.split(":")[2][2:])

try:
    allFirmwareDownloads = json.load(open("firmwareDownloads.json","r"))
except:
    allFirmwareDownloads = []

access = ""
for fn in ("/var/log/nginx/access.log.1", "/var/log/nginx/access.log"):
    try:
        with open(fn) as fp:
            access +=fp.read()
    except Exception as e:
        logging.warning(f"not using {fn}")

try:
    access += open("access.log").read()
except:
    pass

allDownloads = access.strip().split("\n")
allNewFirmwareDownloads = [l for l in allDownloads if l.startswith("fd21:") and "sysupgrade" in l and "manifest" not in l]

for f in allNewFirmwareDownloads:
    if f not in allFirmwareDownloads:
        allFirmwareDownloads.append(f)

json.dump(allFirmwareDownloads,open("firmwareDownloads.json","w"),sort_keys=True, indent=4, separators=(',', ': '))


data = json.load(args.raw)

updateEvents = {}
for download in allFirmwareDownloads:
    try:
        r = LogRecord(download)
    except LogRecordParseError as e:
        logging.debug("error parsing line, skipping: {}".format(e.record_line))
        continue

    mac = ipv62mac(r.ipv6)

    nodes_with_mac = [node for node in data["nodes"] if node["nodeinfo"]["network"]["mac"] == mac]

    if len(nodes_with_mac) == 1:
        d = nodes_with_mac[0]
        currentRelease = d["nodeinfo"]["software"]["firmware"]["release"]
        status = d["online"]
        hostname = d["nodeinfo"]["hostname"]
        hardware_model = getHardwareModelFromEntry(d)
        if currentRelease < r.release:
            if mac in updateEvents:
                updateEvents[mac]["download_count"] += 1
                updateEvents[mac]["last_date"] = r.date
            else:
                updateEvents[mac] = {
                    "mac": mac,
                    "hostname": hostname,
                    "from_release": currentRelease,
                    "to_release": r.release,
                    "segment": r.segment,
                    "model": hardware_model,
                    "status": status,
                    "date": r.date,
                    "last_date": r.date,
                    "download_count": 1
                }
    else:
        # raw.json only contains data from 14 days - do not print message if we encounter older log entries
        download_ts = datetime.datetime.strptime(r.date, "%d/%b/%Y:%H:%M:%S %z]")
        if (datetime.datetime.now(tz=datetime.timezone.utc) - download_ts) < datetime.timedelta(days=14):
            logging.info("%s %s with agent %s seems not to be a node @ %s"%(r.ipv6,mac,r.agent,r.date))

if not args.json:
    for event in updateEvents.values():
        print("%s (%s) %s -> %s Segment %i %s status %s @ %s" % (event["mac"], filterNonAscii(event["hostname"]), event["from_release"], event["to_release"], event["segment"], event["model"], event["status"], event["date"]))
else:
    outJson = {
        "timestamp": time.time(),
        "update_events": list(updateEvents.values())
    }
    json.dump(outJson, args.json)
